﻿using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.IO;
using System.Linq;
using System.Text;
using System.Linq.Dynamic;
using System.Web.Script.Services;
using System.Web.Services;

namespace JData.Services
{
    [WebService(Namespace = "http://tempuri.org/")]
    [WebServiceBinding(ConformsTo = WsiProfiles.BasicProfile1_1)]
    [System.ComponentModel.ToolboxItem(false)]
    [ScriptService]
    public abstract class DataContextService<TDataContext> : WebService
        where TDataContext : ALinq.DataContext
    {

        protected QueryResult Query<T>(int skip, int take, string sorting, string filter, string selector,
                                       bool retrieveTotalRowCount, int totalRowCount)
        {
            var db = Activator.CreateInstance<TDataContext>();
            db.Connection.Open();
            try
            {
                IQueryable q = db.GetTable(typeof(T));

                if (filter != null)
                    q = q.Where(filter);

                if (sorting != null)
                    q = q.OrderBy(sorting);

                int total;
                if (retrieveTotalRowCount && totalRowCount < 0)
                    total = q.Count();
                else
                    total = totalRowCount;

                if (skip > 0)
                    q = q.Skip(skip);

                if (take > 0)
                    q = q.Take(take);

                if (!string.IsNullOrEmpty(selector))
                {
                    selector = selector.Trim();
                    if (!selector.StartsWith("new"))
                        selector = string.Format("new({0})", selector);
                    q = q.Select(selector);
                }
                return new QueryResult { Items = q.Cast<object>().ToArray(), Total = total };
            }
            finally
            {
                db.Connection.Close();
            }
        }

        public ExecuteResult Update<T>(Dictionary<string, object> item) where T : class
        {
            var db = CreateDataContext();
            db.Connection.Open();
            db.Transaction = db.Connection.BeginTransaction();
            try
            {
                var result = new ExecuteResult { AffectedRowCount = Update(db, typeof(T), item) };
                db.Transaction.Commit();
                return result;
            }
            catch (Exception exc)
            {
                db.Transaction.Rollback();
                return new ExecuteResult { AffectedRowCount = 0, Message = exc.Message };
            }
            finally
            {
                db.Connection.Close();
            }
        }

        private static int Update(TDataContext db, Type t, Dictionary<string, object> item)
        {
            var count = 0;
            var metaType = db.Mapping.GetMetaType(t);

            var primaryKeys = metaType.DataMembers.Where(o => o.IsPrimaryKey).ToArray();
            if (primaryKeys.Length == 0)
                throw new Exception("The object type does not contains primary key.");

            var members = t.GetMembers();
            for (var i = 0; i < primaryKeys.Length; i++)
            {
                var primaryKey = primaryKeys[i];
                if (item.Keys.Contains(primaryKey.Member.Name) == false)
                    throw new Exception(string.Format("The item does not contains '{0}' primary key.", primaryKey.Name));
            }

            var values = new List<object>();
            var sb = new StringBuilder();
            sb.AppendLine("Update " + metaType.Table.TableName);
            sb.Append("Set ");

            int x = 0;
            foreach (var key in item.Keys)
            {
                x++;
                var member = members.SingleOrDefault(o => o.Name == key);
                if (member == null)
                    continue;

                var metaMember = metaType.GetDataMember(member);
                if (metaMember.IsPrimaryKey || metaMember.IsDbGenerated || metaMember.IsDiscriminator)
                    continue;

                var value = item[key];
                if (metaMember.IsAssociation)
                {
                    if (metaMember.Type.IsGenericType)
                    {
                        var types = metaMember.Type.GetGenericArguments();
                        Debug.Assert(types.Length == 1);
                        Debug.Assert(item[key] is Array);
                        var children = (object[])item[key];
                        for (var i = 0; i < children.Length; i++)
                        {
                            var subItem = (Dictionary<string, object>)children[i];
                            count = count + Update(db, types[0], subItem);
                        }
                    }
                    else
                    {
                        if (value is Dictionary<string, object>)
                            count = count + Update(db, metaMember.Type, (Dictionary<string, object>)value);
                    }
                    continue;
                }

                if (value != null)
                {
                    sb.Append(metaMember.MappedName + " = {" + values.Count + "}");
                    if (value is DateTime)
                        value = ((DateTime) value).ToLocalTime();

                    values.Add(value);
                }
                else
                {
                    sb.Append(metaMember.MappedName + " = null");
                }

                sb.Append(",");
            }
            sb.Remove(sb.Length - 1, 1);
            sb.AppendLine();

            sb.Append("Where " + primaryKeys[0].MappedName + " = {" + values.Count + "}");
            values.Add(item[primaryKeys[0].Member.Name]);
            for (var i = 1; i < primaryKeys.Length; i++)
            {
                sb.Append(" And " + primaryKeys[i].MappedName + " = {" + values.Count + "}");
                values.Add(item[primaryKeys[i].Member.Name]);
            }
            count = count + db.ExecuteCommand(sb.ToString(), values.ToArray());
            return count;

        }

        public ExecuteResult Delete<T>(Dictionary<string, object> item) where T : class
        {
            var db = CreateDataContext();
            db.Connection.Open();
            db.Transaction = db.Connection.BeginTransaction();
            try
            {
                var result = new ExecuteResult { AffectedRowCount = Delete(db, typeof(T), item) };
                db.Transaction.Commit();
                return result;
            }
            catch (Exception exc)
            {
                db.Transaction.Rollback();
                return new ExecuteResult { AffectedRowCount = 0, Message = exc.Message };
            }
            finally
            {
                db.Connection.Close();
            }
        }

        private static int Delete(TDataContext db, Type t, Dictionary<string, object> item)
        {
            var count = 0;
            var metaType = db.Mapping.GetMetaType(t);

            var primaryKeys = metaType.DataMembers.Where(o => o.IsPrimaryKey).ToArray();
            if (primaryKeys.Length == 0)
                throw new Exception("The object type does not contains primary key.");

            for (var i = 0; i < primaryKeys.Length; i++)
            {
                var primaryKey = primaryKeys[i];
                if (item.Keys.Contains(primaryKey.Member.Name) == false)
                    throw new Exception(string.Format("The item does not contains '{0}' primary key.", primaryKey.Name));
            }

            var values = new List<object>();
            var sb = new StringBuilder();
            sb.AppendLine("Delete From " + metaType.Table.TableName);

            sb.Append("Where " + primaryKeys[0].MappedName + " = {" + values.Count + "}");
            values.Add(item[primaryKeys[0].Member.Name]);
            for (var i = 1; i < primaryKeys.Length; i++)
            {
                sb.Append(" And " + primaryKeys[i].MappedName + " = {" + values.Count + "}");
                values.Add(item[primaryKeys[i].Member.Name]);
            }
            count = count + db.ExecuteCommand(sb.ToString(), values.ToArray());
            return count;

        }

        public ExecuteResult Insert<T>(Dictionary<string, object> item)
        {
            var db = CreateDataContext();
            try
            {
                db.Connection.Open();
                db.Transaction = db.Connection.BeginTransaction();
                var count = Insert(db, typeof(T), item);
                db.Transaction.Commit();
                return new ExecuteResult { AffectedRowCount = count };
            }
            catch (Exception exc)
            {
                db.Transaction.Rollback();
                string message;
                if (exc.InnerException != null)
                    message = exc.InnerException.Message;
                else
                    message = exc.Message;

                return new ExecuteResult { AffectedRowCount = 0, Message = message };
            }
            finally
            {
                db.Connection.Close();
            }
        }

        public int Insert(TDataContext db, Type t, Dictionary<string, object> item)
        {
            var count = 0;
            //var newItem = Activator.CreateInstance(t);
            var metaType = db.Mapping.GetMetaType(t);
            var members = t.GetMembers();

            var fields = new List<KeyValuePair<string, object>>();
            foreach (var key in item.Keys)
            {
                var member = members.SingleOrDefault(o => o.Name == key);
                if (member == null)
                    continue;

                var metaMember = metaType.GetDataMember(member);
                if (metaMember.IsAssociation)
                {
                    continue;
                }
                object value = item[key];
                if (value == null)
                    continue;

                var type = metaMember.Type;
                if (type.IsGenericType && type.GetGenericTypeDefinition() == typeof(Nullable<>))
                {
                    type = type.GetGenericArguments()[0];
                }
                
                if (value is DateTime)
                    value = ((DateTime)value).ToLocalTime();

                if (type != value.GetType())
                    value = Convert.ChangeType(value, type);

               fields.Add(new KeyValuePair<string, object>(metaMember.MappedName, value));
            }

            var insertBuilder = new StringBuilder();
            insertBuilder.AppendLine("INSERT INTO " + metaType.Table.TableName);
            insertBuilder.AppendLine("      (" + string.Join(",", fields.Select(o => o.Key).ToArray()) + ")");
            insertBuilder.Append("Values(");
            for (var i = 0; i < fields.Count; i++)
            {
                insertBuilder.Append("{" + i + "}");
                if (i != fields.Count - 1)
                    insertBuilder.Append(",");
            }
            insertBuilder.Append(")");
            count = count + db.ExecuteCommand(insertBuilder.ToString(), fields.Select(o => o.Value).ToArray());

            var autoColumns = metaType.DataMembers.Where(o => o.IsDbGenerated).Select(o => o.MappedName).ToArray();
            if (autoColumns.Length > 1)
                throw new Exception("The database generated columns is more than 1.");

            if (autoColumns.Length > 0)
            {
                if (db.Mapping.ProviderType.Name == "AccessDbProvider")
                {
                    var command = db.Connection.CreateCommand();
                    command.Transaction = db.Transaction;
                    command.CommandText = "Select @@IDENTITY From " + metaType.Table.TableName;
                    var value = command.ExecuteScalar();
                    Debug.Assert(value != null);
                    item[autoColumns[0]] = value;
                }
                else
                {
                    throw new NotSupportedException(string.Format("The {0} provider is not supported", db.Mapping.ProviderType.Name));
                }

            }

            foreach (var key in item.Keys)
            {
                var member = members.SingleOrDefault(o => o.Name == key);
                if (member == null)
                    continue;

                var metaMember = metaType.GetDataMember(member);
                Debug.Assert(metaMember != null);
                if (!metaMember.IsAssociation)
                    continue;

                var metaAssociation = metaMember.Association;
                if (metaMember.Type.IsGenericType)
                {
                    var types = metaMember.Type.GetGenericArguments();
                    Debug.Assert(types.Length == 1);
                    Debug.Assert(item[key] is Array);
                    var children = (object[])item[key];
                    for (var i = 0; i < children.Length; i++)
                    {
                        var subItem = (Dictionary<string, object>)children[i];
                        Debug.Assert(metaAssociation.ThisKey.Count == metaAssociation.OtherKey.Count);
                        for (var j = 0; j < metaAssociation.ThisKey.Count; j++)
                        {
                            var value = item[metaAssociation.ThisKey[j].Member.Name];
                            Debug.Assert(value != null);
                            subItem[metaAssociation.OtherKey[j].Member.Name] = value;
                        }
                        count = count + Insert(db, types[0], subItem);
                    }
                }
            }

            return count;
        }

        protected TDataContext CreateDataContext()
        {
            var db = Activator.CreateInstance<TDataContext>();
            db.Log = Log;
            return db;
        }

        public TextWriter Log { get; set; }
    }
}
