﻿using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.IO;
using System.Linq;
using System.Linq.Expressions;
using System.Reflection;
using System.Web.Script.Services;
using System.Web.Services;
using System.Linq.Dynamic;
using ALinq.Mapping;

namespace ALinq.Web.Services
{
    public class QueryArgument
    {
        public Type EntityType;
        public string Select;
        public string Order;
        public string Where;
        public int Take;
        public int Skip;

        public QueryArgument()
        {
            this.EntityType = null;
            this.Select = null;
            this.Order = null;
            this.Where = null;
            this.Take = -1;
            this.Skip = -1;
        }
    }

    public class QueryResult
    {
        public object[] Items;
        public int Total = -1;
    }

    public class ExecuteResult
    {
        public int AffectedRowCount;
        public object Message;

        /// <summary>
        /// Used to update the client object members.
        /// </summary>
        public object Members;
    }

    [WebService(Namespace = "http://tempuri.org/")]
    [WebServiceBinding(ConformsTo = WsiProfiles.BasicProfile1_1)]
    [System.ComponentModel.ToolboxItem(false)]
    [ScriptService]
    public abstract class DataContextService<TDataContext> : WebService
        where TDataContext : ALinq.DataContext
    {

        protected QueryResult Query<T>(int skip, int take, string sorting, string filter, string selector,
                                       bool retrieveTotalRowCount, int totalRowCount)
        {
            var db = Activator.CreateInstance<TDataContext>();
            db.Connection.Open();
            try
            {
                IQueryable q = db.GetTable(typeof(T));

                if (filter != null)
                    q = q.Where(filter);

                if (sorting != null)
                    q = q.OrderBy(sorting);

                int total;
                if (retrieveTotalRowCount && totalRowCount < 0)
                    total = q.Count();
                else
                    total = totalRowCount;

                if (skip > 0)
                    q = q.Skip(skip);

                if (take > 0)
                    q = q.Take(take);

                if (!string.IsNullOrEmpty(selector))
                {
                    selector = selector.Trim();
                    if (!selector.StartsWith("new"))
                        selector = string.Format("new({0})", selector);
                    q = q.Select(selector);
                }
                return new QueryResult { Items = q.Cast<object>().ToArray(), Total = total };
            }
            finally
            {
                db.Connection.Close();
            }
        }

        public ExecuteResult Update<T>(Dictionary<string, object> item) where T : class
        {
            var db = CreateDataContext();
            db.Connection.Open();
            db.Transaction = db.Connection.BeginTransaction();
            try
            {
                var result = new ExecuteResult { AffectedRowCount = Update(db, typeof(T), item) };
                db.Transaction.Commit();
                return result;
            }
            catch (Exception exc)
            {
                db.Transaction.Rollback();
                return new ExecuteResult { AffectedRowCount = 0, Message = exc.Message };
            }
            finally
            {
                db.Connection.Close();
            }
        }

        private static int Update(TDataContext db, Type t, Dictionary<string, object> item)
        {
            var count = 0;
            var metaType = db.Mapping.GetMetaType(t);

            var primaryKeys = metaType.DataMembers.Where(o => o.IsPrimaryKey).ToArray();
            if (primaryKeys.Length == 0)
                throw new Exception("The object type does not contains primary key.");

            var members = t.GetMembers();
            for (var i = 0; i < primaryKeys.Length; i++)
            {
                var primaryKey = primaryKeys[i];
                if (item.Keys.Contains(primaryKey.Member.Name) == false)
                    throw new Exception(string.Format("The item does not contains '{0}' primary key.", primaryKey.Name));
            }

            var parameter = Expression.Parameter(t, "o");
            var property = Expression.Property(parameter, primaryKeys[0].Member.Name);
            var constant = Expression.Constant(item[primaryKeys[0].Member.Name]);
            var binary = Expression.Equal(property, constant);
            for (var i = 1; i < primaryKeys.Length; i++)
            {
                property = Expression.Property(parameter, primaryKeys[i].Member.Name);
                constant = Expression.Constant(item[primaryKeys[i].Member.Name]);
                binary = Expression.And(binary, Expression.Equal(property, constant));
            }
            var predicate = Expression.Lambda(binary, parameter);

            parameter = Expression.Parameter(t, "o");
            var newExp = Expression.New(t);
            var binds = new List<MemberBinding>();
            foreach (var primaryKey in primaryKeys)
            {
                var member = primaryKey.Member;
                var bind = Expression.Bind(member, Expression.MakeMemberAccess(parameter, member));
                binds.Add(bind);
            }
            foreach (var key in item.Keys)
            {
                var member = members.SingleOrDefault(o => o.Name == key);
                if (member == null)
                    continue;

                var metaMember = metaType.GetDataMember(member);
                if (metaMember.IsPrimaryKey || metaMember.IsDbGenerated || metaMember.IsDiscriminator)
                    continue;

                if (metaMember.IsAssociation)
                {
                    if (metaMember.Type.IsGenericType)
                    {
                        var types = metaMember.Type.GetGenericArguments();
                        Debug.Assert(types.Length == 1);
                        Debug.Assert(item[key] is Array);
                        var children = (object[])item[key];
                        for (var i = 0; i < children.Length; i++)
                        {
                            var subItem = (Dictionary<string, object>)children[i];
                            count = count + Update(db, types[0], subItem);
                        }
                    }
                    else
                    {
                        if (item[key] is Dictionary<string, object>)
                            count = count + Update(db, metaMember.Type, (Dictionary<string, object>)item[key]);
                    }
                    continue;
                }

                Type type;
                if (member is PropertyInfo)
                    type = ((PropertyInfo)member).PropertyType;
                else
                    type = ((FieldInfo)member).FieldType;

                if (metaMember.CanBeNull && type.IsValueType && !type.IsGenericType)
                    type = typeof(Nullable<>).MakeGenericType(type);

                MemberAssignment bind;
                if (item[key] != null && item[key].GetType() == type)
                    bind = Expression.Bind(member, Expression.Constant(item[key]));
                else
                    bind = Expression.Bind(member, Expression.Convert(Expression.Constant(item[key]), type));
                binds.Add(bind);
            }
            var memberInit = Expression.MemberInit(newExp, binds);
            var entity = Expression.Lambda(memberInit, parameter);
            var table = db.GetTable(t);
            count = count + table.Update(t, t, entity, predicate);
            return count;

        }

        public ExecuteResult Delete<T>(Dictionary<string, object> item) where T : class
        {
            var db = CreateDataContext();
            db.Connection.Open();
            db.Transaction = db.Connection.BeginTransaction();
            try
            {
                var result = new ExecuteResult { AffectedRowCount = Delete(db, typeof(T), item) };
                db.Transaction.Commit();
                return result;
            }
            catch (Exception exc)
            {
                db.Transaction.Rollback();
                return new ExecuteResult { AffectedRowCount = 0, Message = exc.Message };
            }
            finally
            {
                db.Connection.Close();
            }
        }

        private static int Delete(TDataContext db, Type t, Dictionary<string, object> item)
        {
            var count = 0;
            var metaType = db.Mapping.GetMetaType(t);

            var primaryKeys = metaType.DataMembers.Where(o => o.IsPrimaryKey).ToArray();
            if (primaryKeys.Length == 0)
                throw new Exception("The object type does not contains primary key.");

            for (var i = 0; i < primaryKeys.Length; i++)
            {
                var primaryKey = primaryKeys[i];
                if (item.Keys.Contains(primaryKey.Member.Name) == false)
                    throw new Exception(string.Format("The item does not contains '{0}' primary key.", primaryKey.Name));
            }

            var members = t.GetMembers();
            foreach (var key in item.Keys)
            {
                var member = members.SingleOrDefault(o => o.Name == key);
                if (member == null)
                    continue;

                var metaMember = metaType.GetDataMember(member);
                if (metaMember.IsPrimaryKey || metaMember.IsDbGenerated || metaMember.IsDiscriminator)
                    continue;

                if (metaMember.IsAssociation)
                {
                    if (metaMember.Type.IsGenericType)
                    {
                        var types = metaMember.Type.GetGenericArguments();
                        Debug.Assert(types.Length == 1);
                        Debug.Assert(item[key] is Array);
                        var children = (object[])item[key];
                        for (var i = 0; i < children.Length; i++)
                        {
                            var subItem = (Dictionary<string, object>)children[i];
                            count = count + Delete(db, types[0], subItem);
                        }
                    }
                    else
                    {
                        count = count + Delete(db, metaMember.Type, (Dictionary<string, object>)item[key]);
                    }
                }
            }

            var parameter = Expression.Parameter(t, "o");
            var property = Expression.Property(parameter, primaryKeys[0].Member.Name);
            var constant = Expression.Constant(item[primaryKeys[0].Member.Name]);
            var binary = Expression.Equal(property, constant);
            for (var i = 1; i < primaryKeys.Length; i++)
            {
                property = Expression.Property(parameter, primaryKeys[i].Member.Name);
                constant = Expression.Constant(item[primaryKeys[i].Member.Name]);
                binary = Expression.And(binary, Expression.Equal(property, constant));
            }
            var predicate = Expression.Lambda(binary, parameter);
            var table = db.GetTable(t);
            count = count + table.Delete(t, predicate);
            return count;
        }

        public ExecuteResult Insert<T>(Dictionary<string, object> item)
        {
            var db = CreateDataContext();
            try
            {
                db.Connection.Open();
                db.Transaction = db.Connection.BeginTransaction();
                var count = Insert(db, typeof(T), item);
                db.Transaction.Commit();
                return new ExecuteResult { AffectedRowCount = count };
            }
            catch (Exception exc)
            {
                db.Transaction.Rollback();
                return new ExecuteResult { AffectedRowCount = 0, Message = exc.InnerException.Message };
            }
            finally
            {
                db.Connection.Close();
            }
        }

        public int Insert(TDataContext db, Type t, Dictionary<string, object> item)
        {
            var count = 0;
            var newItem = Activator.CreateInstance(t);
            var metaType = db.Mapping.GetMetaType(t);
            var members = t.GetMembers();
            foreach (var key in item.Keys)
            {
                var member = members.SingleOrDefault(o => o.Name == key);
                if (member == null)
                    continue;

                var metaMember = metaType.GetDataMember(member);
                if (metaMember.IsAssociation)
                {
                    continue;
                }
                object value = item[key];
                var type = metaMember.Type;
                if (type.IsGenericType && type.GetGenericTypeDefinition() == typeof(Nullable<>))
                {
                    type = type.GetGenericArguments()[0];
                }
                if (value != null && (type != value.GetType()))
                    value = Convert.ChangeType(value, type);
                metaMember.MemberAccessor.SetBoxedValue(ref newItem, value);
            }
            var table = db.GetTable(t);
            var insertMethod = table.GetType().GetMethod("Insert", new[] { t });
            Debug.Assert(insertMethod != null);
            insertMethod.Invoke(table, new[] { newItem });

            count = count + 1;

            foreach (var key in item.Keys)
            {
                var member = members.SingleOrDefault(o => o.Name == key);
                if (member == null)
                    continue;

                var metaMember = metaType.GetDataMember(member);
                Debug.Assert(metaMember != null);
                if (!metaMember.IsAssociation)
                    continue;

                var metaAssociation = metaMember.Association;
                if (metaMember.Type.IsGenericType)
                {
                    var types = metaMember.Type.GetGenericArguments();
                    Debug.Assert(types.Length == 1);
                    Debug.Assert(item[key] is Array);
                    var children = (object[])item[key];
                    for (var i = 0; i < children.Length; i++)
                    {
                        var subItem = (Dictionary<string, object>)children[i];
                        Debug.Assert(metaAssociation.ThisKey.Count == metaAssociation.OtherKey.Count);
                        for (var j = 0; j < metaAssociation.ThisKey.Count; j++)
                        {
                            var m = metaAssociation.ThisKey[j].MemberAccessor;
                            var value = m.GetBoxedValue(newItem);
                            Debug.Assert(value != null);
                            subItem[metaAssociation.OtherKey[j].Member.Name] = value;
                            //Insert(db, types[0], subItem);
                        }
                        count = count + Insert(db, types[0], subItem);
                    }
                }
            }

            return count;
        }

        protected TDataContext CreateDataContext()
        {
            var db = Activator.CreateInstance<TDataContext>();
            db.Log = Log;
            return db;
        }

        public TextWriter Log { get; set; }
    }
}
